#ifndef _SHAPEFILE_H_INCLUDED
#define _SHAPEFILE_H_INCLUDED

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <ctype.h>

#ifdef USE_DBMALLOC
#include <dbmalloc.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/************************************************************************/
/*                        Configuration options.                        */
/************************************************************************/

/* -------------------------------------------------------------------- */
/*      Should the DBFReadStringAttribute() strip leading and           */
/*      trailing white space?                                           */
/* -------------------------------------------------------------------- */
#define TRIM_DBF_WHITESPACE

/* -------------------------------------------------------------------- */
/*      Should we write measure values to the Multipatch object?        */
/*      Reportedly ArcView crashes if we do write it, so for now it     */
/*      is disabled.                                                    */
/* -------------------------------------------------------------------- */
#define DISABLE_MULTIPATCH_MEASURE

/* -------------------------------------------------------------------- */
/*      SHPAPI_CALL                                                     */
/*                                                                      */
/*      The following two macros are present to allow forcing           */
/*      various calling conventions on the Shapelib API.                */
/*                                                                      */
/*      To force __stdcall conventions (needed to call Shapelib         */
/*      from Visual Basic and/or Dephi I believe) the makefile could    */
/*      be modified to define:                                          */
/*                                                                      */
/*        /DSHPAPI_CALL=__stdcall                                       */
/*                                                                      */
/*      If it is desired to force export of the Shapelib API without    */
/*      using the shapelib.def file, use the following definition.      */
/*                                                                      */
/*        /DSHAPELIB_DLLEXPORT                                          */
/*                                                                      */
/*      To get both at once it will be necessary to hack this           */
/*      include file to define:                                         */
/*                                                                      */
/*        #define SHPAPI_CALL __declspec(dllexport) __stdcall           */
/*        #define SHPAPI_CALL1 __declspec(dllexport) * __stdcall        */
/*                                                                      */
/*      The complexity of the situtation is partly caused by the        */
/*      peculiar requirement of Visual C++ that __stdcall appear        */
/*      after any "*"'s in the return value of a function while the     */
/*      __declspec(dllexport) must appear before them.                  */
/* -------------------------------------------------------------------- */

#ifdef SHAPELIB_DLLEXPORT
#  define SHPAPI_CALL __declspec(dllexport)
#  define SHPAPI_CALL1(x)  __declspec(dllexport) x
#endif

#ifndef SHPAPI_CALL
#  define SHPAPI_CALL
#endif

#ifndef SHPAPI_CALL1
#  define SHPAPI_CALL1(x)      x SHPAPI_CALL
#endif
    
/************************************************************************/
/*                             SHP Support.                             */
/************************************************************************/
typedef	struct
{
	int				nShapeType;				/* SHPT_* */

	int				nFileSize;				/* SHP file */

	int     		nRecords;
	int				nMaxRecords;
	int				*panRecOffset;
	int				*panRecSize;

	double			adBoundsMin[4];
	double			adBoundsMax[4];

	int				bUpdated;

	unsigned char 	*pabyRec;
	int         	nBufSize;
} SHPInfo;

typedef SHPInfo* 	SHPHandle;

/* -------------------------------------------------------------------- */
/*      Shape types (nSHPType)                                          */
/* -------------------------------------------------------------------- */
typedef enum
{
	SHPT_NULL = 0,
	SHPT_POINT = 	1,
	SHPT_ARC = 	3,
	SHPT_POLYGON = 	5,
	SHPT_MULTIPOINT = 	8,
	SHPT_POINTZ = 	11,
	SHPT_ARCZ = 	13,
	SHPT_POLYGONZ = 	15,
	SHPT_MULTIPOINTZ =  18,
	SHPT_POINTM = 	21,
	SHPT_ARCM = 	23,
	SHPT_POLYGONM = 	25,
	SHPT_MULTIPOINTM =  28,
	SHPT_MULTIPATCH =  31,
} TShpType;


/* -------------------------------------------------------------------- */
/*      Part types - everything but SHPT_MULTIPATCH just uses           */
/*      SHPP_RING.                                                      */
/* -------------------------------------------------------------------- */

typedef enum
{
SHPP_TRISTRIP = 0,
SHPP_TRIFAN = 	1,
SHPP_OUTERRING = 	2,
SHPP_INNERRING = 	3,
SHPP_FIRSTRING = 	4,
SHPP_RING = 	5,
} TPartType;

/* -------------------------------------------------------------------- */
/*      SHPObject - represents on shape (without attributes) read       */
/*      from the .shp file.                                             */
/* -------------------------------------------------------------------- */
typedef struct
{
    int		nSHPType;

    int		nShapeId; /* -1 is unknown/unassigned */

    int		nParts;
    int		*panPartStart;
    int		*panPartType;
    
    int		nVertices;
    double	*padfX;
    double	*padfY;
    double	*padfZ;
    double	*padfM;

    double	dfXMin;
    double	dfYMin;
    double	dfZMin;
    double	dfMMin;

    double	dfXMax;
    double	dfYMax;
    double	dfZMax;
    double	dfMMax;
} SHPObject;

/* -------------------------------------------------------------------- */
/*      SHP API Prototypes                                              */
/* -------------------------------------------------------------------- */
SHPHandle SHPAPI_CALL 		SHPOpen( const char * pszShapeFile, const char * pszAccess );
SHPHandle SHPAPI_CALL 		SHPCreate( const char * pszShapeFile, int nShapeType );
void SHPAPI_CALL 			SHPGetInfo( SHPHandle hSHP, int * pnEntities, TShpType *pnShapeType, double * padfMinBound, double * padfMaxBound );

SHPObject SHPAPI_CALL1(*)	SHPReadObject( SHPHandle hSHP, int iShape );
void SHPAPI_CALL 			SHPSetMaxRecords(SHPHandle psSHP, int nNewMaxRecords);
int SHPAPI_CALL             SHPWriteObject( SHPHandle hSHP, int iShape, SHPObject * psObject );
int SHPAPI_CALL 			SHPWritePoint(SHPHandle psSHP, double X, double Y);
int SHPAPI_CALL 			SHPWriteElement(SHPHandle psSHP, double X1, double Y1, double X2, double Y2, double X3, double Y3);
int SHPAPI_CALL 			SHPWritePolygon(SHPHandle psSHP, SHPObject* psObject);

void SHPAPI_CALL            SHPDestroyObject( SHPObject * psObject );
void SHPAPI_CALL            SHPComputeExtents( SHPObject * psObject );
SHPObject SHPAPI_CALL1(*)   SHPCreateObject( int nSHPType, int nShapeId, int nParts, int * panPartStart, int * panPartType, int nVertices, double * padfX, double * padfY, double * padfZ, double * padfM );
SHPObject SHPAPI_CALL1(*)   SHPCreateSimpleObject( int nSHPType, int nVertices, double * padfX, double * padfY, double * padfZ );

int SHPAPI_CALL            	SHPRewindObject( SHPHandle hSHP, SHPObject * psObject );

void SHPAPI_CALL            SHPClose( SHPHandle hSHP );

const char SHPAPI_CALL1(*)  SHPTypeName( int nSHPType );
const char SHPAPI_CALL1(*)  SHPPartTypeName( int nPartType );

/* -------------------------------------------------------------------- */
/*      Shape quadtree indexing API.                                    */
/* -------------------------------------------------------------------- */

/* this can be two or four for binary or quad tree */
#define MAX_SUBNODE	4

typedef struct shape_tree_node
{
    /* region covered by this node */
    double	adfBoundsMin[4];
    double	adfBoundsMax[4];

    /* list of shapes stored at this node.  The papsShapeObj pointers
       or the whole list can be NULL */
    int		nShapeCount;
    int		*panShapeIds;
    SHPObject   **papsShapeObj;

    int		nSubNodes;
    struct shape_tree_node *apsSubNode[MAX_SUBNODE];
    
} SHPTreeNode;

typedef struct
{
    SHPHandle   hSHP;
    
    int		nMaxDepth;
    int		nDimension;
    
    SHPTreeNode	*psRoot;
} SHPTree;

SHPTree SHPAPI_CALL1(*) SHPCreateTree( SHPHandle hSHP, int nDimension, int nMaxDepth, double *padfBoundsMin, double *padfBoundsMax );
void    SHPAPI_CALL     SHPDestroyTree( SHPTree * hTree );

int	SHPAPI_CALL         SHPWriteTree( SHPTree *hTree, const char * pszFilename );
SHPTree SHPAPI_CALL     SHPReadTree( const char * pszFilename );

int	SHPAPI_CALL         SHPTreeAddObject( SHPTree * hTree, SHPObject * psObject );
int	SHPAPI_CALL         SHPTreeAddShapeId( SHPTree * hTree, SHPObject * psObject );
int	SHPAPI_CALL         SHPTreeRemoveShapeId( SHPTree * hTree, int nShapeId );

void 	SHPAPI_CALL     SHPTreeTrimExtraNodes( SHPTree * hTree );

int    SHPAPI_CALL1(*)  SHPTreeFindLikelyShapes( SHPTree * hTree, double * padfBoundsMin, double * padfBoundsMax, int * );
int     SHPAPI_CALL 	SHPCheckBoundsOverlap( double *, double *, double *, double *, int );

/************************************************************************/
/*                             DBF Support.                             */
/************************************************************************/
typedef	struct
{
    FILE	*fp;

    int         nRecords;

    int		nRecordLength;
    int		nHeaderLength;
    int		nFields;
    int		*panFieldOffset;
    int		*panFieldSize;
    int		*panFieldDecimals;
    char	*pachFieldType;

    char	*pszHeader;

    int		nCurrentRecord;
    int		bCurrentRecordModified;
    char	*pszCurrentRecord;
    
    int		bNoHeader;
    int		bUpdated;
} DBFInfo;

typedef DBFInfo *DBFHandle;

typedef enum {
  FTString,
  FTInteger,
  FTDouble,
  FTLogical,
  FTInvalid
} DBFFieldType;

#define XBASE_FLDHDR_SZ       32

DBFHandle SHPAPI_CALL		DBFOpen( const char * pszDBFFile, const char * pszAccess );
DBFHandle SHPAPI_CALL   	DBFCreate( const char * pszDBFFile );

int	SHPAPI_CALL         	DBFGetFieldCount( DBFHandle psDBF );
int	SHPAPI_CALL         	DBFGetRecordCount( DBFHandle psDBF );
int	SHPAPI_CALL         	DBFAddField( DBFHandle hDBF, const char * pszFieldName, DBFFieldType eType, int nWidth, int nDecimals );

DBFFieldType SHPAPI_CALL    DBFGetFieldInfo( DBFHandle psDBF, int iField, char * pszFieldName, int * pnWidth, int * pnDecimals );

int SHPAPI_CALL      		DBFGetFieldIndex(DBFHandle psDBF, const char *pszFieldName);

int 	SHPAPI_CALL         DBFReadIntegerAttribute( DBFHandle hDBF, int iShape, int iField );
double 	SHPAPI_CALL         DBFReadDoubleAttribute( DBFHandle hDBF, int iShape, int iField );
const char SHPAPI_CALL1(*)  DBFReadStringAttribute( DBFHandle hDBF, int iShape, int iField );
const char SHPAPI_CALL1(*)  DBFReadLogicalAttribute( DBFHandle hDBF, int iShape, int iField );
int     SHPAPI_CALL         DBFIsAttributeNULL( DBFHandle hDBF, int iShape, int iField );

int SHPAPI_CALL             DBFWriteIntegerAttribute( DBFHandle hDBF, int iShape, int iField, int nFieldValue );
int SHPAPI_CALL	  			DBFWriteDoubleAttribute( DBFHandle hDBF, int iShape, int iField, double dFieldValue );
int SHPAPI_CALL      		DBFWriteStringAttribute( DBFHandle hDBF, int iShape, int iField, const char * pszFieldValue );
int SHPAPI_CALL	 			DBFWriteNULLAttribute( DBFHandle hDBF, int iShape, int iField );

int SHPAPI_CALL             DBFWriteLogicalAttribute( DBFHandle hDBF, int iShape, int iField, const char lFieldValue);
int SHPAPI_CALL				DBFWriteAttributeDirectly(DBFHandle psDBF, int hEntity, int iField, void * pValue );
const char SHPAPI_CALL1(*)  DBFReadTuple(DBFHandle psDBF, int hEntity );
int SHPAPI_CALL             DBFWriteTuple(DBFHandle psDBF, int hEntity, void * pRawTuple );

DBFHandle SHPAPI_CALL       DBFCloneEmpty(DBFHandle psDBF, const char * pszFilename );

void	SHPAPI_CALL         DBFClose( DBFHandle hDBF );
char    SHPAPI_CALL        	DBFGetNativeFieldType( DBFHandle hDBF, int iField );

void	TestShapeFileLib(void);

#ifdef __cplusplus
}
#endif

#endif /* ndef _SHAPEFILE_H_INCLUDED */
